@@ -4,10 +4,10 @@ import torch.nn as nn
 from torch.nn import Identity
 import torch.nn.functional as F
 from torch.autograd import Function
-from functools import partial, reduce
+from functools import partial, reduce, wraps
 from itertools import chain
 from operator import mul
-from reformer_pytorch.reversible import ReversibleBlock, ReversibleSequence
+from reformer_pytorch.reversible import ReversibleSequence
 
 #constants
 
@@ -39,8 +39,15 @@ def chunked_sum(tensor, chunks=1):
     summed_tensors = [c.sum(dim=-1) for c in tensor.chunk(chunks, dim=0)]
     return torch.cat(summed_tensors, dim=0).reshape(orig_size)
 
+def default(val, default_val):
+    return default_val if val is None else val
+
+def max_neg_value(tensor):
+    return -torch.finfo(tensor.dtype).max
+
 def cache_fn(f):
     cache = None
+    @wraps(f)
     def cached_fn(*args, **kwargs):
         nonlocal cache
         if cache is not None:
@@ -49,11 +56,25 @@ def cache_fn(f):
         return cache
     return cached_fn
 
-def default(val, default_val):
-    return default_val if val is None else val
+def cache_method_decorator(cache_attr, cache_key, execute_in_cache = False):
+    def inner_fn(fn):
+        @wraps(fn)
+        def wrapper(self, *args, namespace=None, get_from_cache=False, set_cache=True, **kwargs):
+            namespace_str = str(default(namespace, ''))
+            _cache = getattr(self, cache_attr)
+            _keyname = f'{cache_key}:{namespace_str}'
 
-def max_neg_value(tensor):
-    return -torch.finfo(tensor.dtype).max
+            if get_from_cache:
+                if execute_in_cache:
+                    fn(self, *args, **kwargs)
+                val = _cache[_keyname]
+            else:
+                val = fn(self, *args, **kwargs)
+                if set_cache:
+                    setattr(self, cache_attr, {**_cache, **{_keyname: val}})
+            return val
+        return wrapper
+    return inner_fn
 
 # helper classes
 
@@ -98,9 +119,9 @@ class Chunk(nn.Module):
         self.chunks = chunks
         self.fn = fn
 
-    def forward(self, x):
+    def forward(self, x, **kwargs):
         chunks = x.chunk(self.chunks, dim = self.dim)
-        return torch.cat([self.fn(c) for c in chunks], dim = self.dim)
+        return torch.cat([self.fn(c, **kwargs) for c in chunks], dim = self.dim)
 
 # LSH attention as described in https://openreview.net/pdf?id=rkgNKkHtvB
 # adapted from trax, stripped to what paper said needed to work
@@ -112,7 +133,6 @@ class LSHAttention(nn.Module):
                   dropout = 0.,
                   bucket_size = 64,
                   n_hashes = 8,
-                  add_local_attn_hash = False,
                   causal = False,
                   allow_duplicate_attention = True,
                   attend_across_buckets = True,
@@ -135,7 +155,6 @@ class LSHAttention(nn.Module):
         self.bucket_size = bucket_size
 
         self.n_hashes = n_hashes
-        self.add_local_attn_hash = add_local_attn_hash
 
         self._allow_duplicate_attention = allow_duplicate_attention
         self._attend_across_buckets = attend_across_buckets
@@ -145,6 +164,10 @@ class LSHAttention(nn.Module):
         # will expend extra computation to return attention matrix
         self._return_attn = return_attn
 
+        # cache buckets for reversible network, reported by authors to make Reformer work at depth
+        self._cache = {}
+
+    @cache_method_decorator('_cache', 'buckets', execute_in_cache=True)
     def hash_vectors(self, n_buckets, vecs):
         batch_size = vecs.shape[0]
         device = vecs.device
@@ -191,26 +214,22 @@ class LSHAttention(nn.Module):
 
         return buckets
 
-    def forward(self, qk, v, query_len = None, input_mask = None, input_attn_mask = None):
-        batch_size, seqlen, dim = qk.shape
+    def forward(self, qk, v, query_len = None, input_mask = None, input_attn_mask = None, **kwargs):
+        batch_size, seqlen, dim, device = *qk.shape, qk.device
+
+        query_len = default(query_len, seqlen)
+        is_reverse = kwargs.pop('_reverse', False)
+        depth = kwargs.pop('_depth', None)
 
         assert seqlen % (self.bucket_size * 2) == 0, f'Sequence length ({seqlen}) needs to be divisible by target bucket size  x 2 - {self.bucket_size * 2}'
 
-        query_len = default(query_len, seqlen)
-        device = qk.device
-
         n_buckets = seqlen // self.bucket_size
-
-        buckets = self.hash_vectors(n_buckets, qk)
+        buckets = self.hash_vectors(n_buckets, qk, namespace=depth, get_from_cache=is_reverse, set_cache=self.training)
 
         # We use the same vector as both a query and a key.
         assert int(buckets.shape[1]) == self.n_hashes * seqlen
 
-        if self.add_local_attn_hash:
-            local_buckets = torch.full((batch_size, seqlen), n_buckets, device=device, dtype=torch.long)
-            buckets = torch.cat((buckets, local_buckets), dim=1)
-
-        total_hashes = self.n_hashes + int(self.add_local_attn_hash)
+        total_hashes = self.n_hashes
 
         ticker = torch.arange(total_hashes * seqlen, device=device).unsqueeze(0).expand_as(buckets)
         buckets_and_t = seqlen * buckets + (ticker % seqlen)
@@ -278,7 +297,9 @@ class LSHAttention(nn.Module):
 
         # Causal masking
         if self.causal:
-            mask = bq_t[:, :, :, None] < bkv_t[:, :, None, :].clamp(max=query_len - 1)
+            mask = bq_t[:, :, :, None] < bkv_t[:, :, None, :]
+            if seqlen > query_len:
+                mask = mask & (bkv_t[:, :, None, :] < query_len)
             dots.masked_fill_(mask, masked_value)
             del mask
 
@@ -386,7 +407,7 @@ class FullQKAttention(nn.Module):
         self.causal = causal
         self.dropout = nn.Dropout(dropout)
 
-    def forward(self, qk, v, query_len = None, input_mask = None, input_attn_mask = None):
+    def forward(self, qk, v, query_len = None, input_mask = None, input_attn_mask = None, **kwargs):
         b, seq_len, dim = qk.shape
         query_len = default(query_len, seq_len)
         t = query_len
@@ -426,7 +447,7 @@ class FullQKAttention(nn.Module):
 # Shared qk attention, using either full or LSH attention
 
 class LSHSelfAttention(nn.Module):
-    def __init__(self, dim, heads = 8, bucket_size = 64, n_hashes = 8, add_local_attn_hash = False, causal = False, attn_chunks = 1, random_rotations_per_head = False, attend_across_buckets = True, allow_duplicate_attention = True, num_mem_kv = 0, one_value_head = False, use_full_attn = False, full_attn_thres = None, return_attn = False, post_attn_dropout = 0., dropout = 0., **kwargs):
+    def __init__(self, dim, heads = 8, bucket_size = 64, n_hashes = 8, causal = False, attn_chunks = 1, random_rotations_per_head = False, attend_across_buckets = True, allow_duplicate_attention = True, num_mem_kv = 0, one_value_head = False, use_full_attn = False, full_attn_thres = None, return_attn = False, post_attn_dropout = 0., dropout = 0., **kwargs):
         super().__init__()
         assert dim % heads == 0, 'dimensions must be divisible by number of heads'
 
@@ -442,7 +463,7 @@ class LSHSelfAttention(nn.Module):
         self.to_out = nn.Linear(dim, dim)
 
         self.bucket_size = bucket_size
-        self.lsh_attn = LSHAttention(bucket_size=bucket_size, n_hashes=n_hashes, add_local_attn_hash = add_local_attn_hash, causal=causal, random_rotations_per_head=random_rotations_per_head, attend_across_buckets = attend_across_buckets,  allow_duplicate_attention = allow_duplicate_attention, return_attn = return_attn, dropout = dropout, **kwargs)
+        self.lsh_attn = LSHAttention(bucket_size=bucket_size, n_hashes=n_hashes, causal=causal, random_rotations_per_head=random_rotations_per_head, attend_across_buckets = attend_across_buckets,  allow_duplicate_attention = allow_duplicate_attention, return_attn = return_attn, dropout = dropout, **kwargs)
         self.full_attn = FullQKAttention(causal = causal, dropout = dropout)
         self.post_attn_dropout = nn.Dropout(post_attn_dropout)
 
@@ -454,7 +475,7 @@ class LSHSelfAttention(nn.Module):
 
         self.callback = None
 
-    def forward(self, x, keys = None, input_mask = None, input_attn_mask = None, context_mask = None):
+    def forward(self, x, keys = None, input_mask = None, input_attn_mask = None, context_mask = None, **kwargs):
         device, dtype = x.device, x.dtype
         b, t, e, h, m = *x.shape, self.heads, self.num_mem_kv
 
@@ -496,7 +517,7 @@ class LSHSelfAttention(nn.Module):
             masks['input_attn_mask'] = input_attn_mask
 
         attn_fn = self.lsh_attn if not use_full_attn else self.full_attn
-        partial_attn_fn = partial(attn_fn, query_len = t)
+        partial_attn_fn = partial(attn_fn, query_len = t, **kwargs)
         attn_fn_in_chunks = process_inputs_chunk(partial_attn_fn, chunks = self.attn_chunks)
 
         out, attn, buckets = attn_fn_in_chunks(qk, v, **masks)
@@ -517,18 +538,27 @@ class GELU_(nn.Module):
 GELU = nn.GELU if hasattr(nn, 'GELU') else GELU_
 
 class FeedForward(nn.Module):
-    def __init__(self, dim, mult = 4, dropout = 0., activation = None):
+    def __init__(self, dim, mult = 4, dropout = 0., activation = None, glu = False):
         super().__init__()
         activation = default(activation, GELU)
 
-        self.net = nn.Sequential(
-            nn.Linear(dim, dim * mult),
-            activation(),
-            nn.Dropout(dropout),
-            nn.Linear(dim * mult, dim))
+        self.glu = glu
+        self.w1 = nn.Linear(dim, dim * mult * (2 if glu else 1))
+        self.act = activation()
+        self.dropout = nn.Dropout(dropout)
+        self.w2 = nn.Linear(dim * mult, dim)
 
-    def forward(self, x):
-        return self.net(x)
+    def forward(self, x, **kwargs):
+        if not self.glu:
+            x = self.w1(x)
+            x = self.act(x)
+        else:
+            x, v = self.w1(x).chunk(2, dim=-1)
+            x = self.act(x) * v
+
+        x = self.dropout(x)
+        x = self.w2(x)
+        return x
 
 # positional embeddings
 
@@ -590,7 +620,7 @@ class AxialPositionalEncoding(nn.Module):
 # reformer lm
 
 class Reformer(nn.Module):
-    def __init__(self, dim, depth, max_seq_len, heads = 8, bucket_size = 64, n_hashes = 8, add_local_attn_hash = False, ff_chunks = 100, attn_chunks = None, causal = False, weight_tie = False, lsh_dropout = 0., ff_dropout = 0., ff_activation = None, ff_mult = 4, post_attn_dropout = 0., layer_dropout = 0., lsh_attend_across_buckets = True, lsh_allow_duplicate_attention = True, random_rotations_per_head = False, twin_attention = False, use_scale_norm = False, use_full_attn = False, full_attn_thres = 0, reverse_thres = 0, num_mem_kv = 0, one_value_head = False):
+    def __init__(self, dim, depth, max_seq_len, heads = 8, bucket_size = 64, n_hashes = 8, ff_chunks = 100, attn_chunks = None, causal = False, weight_tie = False, lsh_dropout = 0., ff_dropout = 0., ff_activation = None, ff_mult = 4, ff_glu = False, post_attn_dropout = 0., layer_dropout = 0., lsh_attend_across_buckets = True, lsh_allow_duplicate_attention = True, random_rotations_per_head = False, twin_attention = False, use_scale_norm = False, use_full_attn = False, full_attn_thres = 0, reverse_thres = 0, num_mem_kv = 0, one_value_head = False):
         super().__init__()
         self.dim = dim
         self.depth = depth
@@ -601,8 +631,8 @@ class Reformer(nn.Module):
         self.twin_attention = twin_attention
         self.full_attn_thres = full_attn_thres
 
-        get_attn = lambda: LSHSelfAttention(dim, heads, bucket_size, n_hashes, add_local_attn_hash = add_local_attn_hash, causal = causal, dropout = lsh_dropout, post_attn_dropout = post_attn_dropout, attn_chunks = attn_chunks, allow_duplicate_attention = lsh_allow_duplicate_attention, attend_across_buckets = lsh_attend_across_buckets, random_rotations_per_head = random_rotations_per_head, num_mem_kv = num_mem_kv, use_full_attn = use_full_attn, full_attn_thres = full_attn_thres, one_value_head = one_value_head)
-        get_ff = lambda: FeedForward(dim, dropout = ff_dropout, activation = ff_activation, mult = ff_mult)
+        get_attn = lambda: LSHSelfAttention(dim, heads, bucket_size, n_hashes, causal = causal, dropout = lsh_dropout, post_attn_dropout = post_attn_dropout, attn_chunks = attn_chunks, allow_duplicate_attention = lsh_allow_duplicate_attention, attend_across_buckets = lsh_attend_across_buckets, random_rotations_per_head = random_rotations_per_head, num_mem_kv = num_mem_kv, use_full_attn = use_full_attn, full_attn_thres = full_attn_thres, one_value_head = one_value_head)
+        get_ff = lambda: FeedForward(dim, dropout = ff_dropout, activation = ff_activation, mult = ff_mult, glu = ff_glu)
 
         if weight_tie:
             get_attn = cache_fn(get_attn)
@@ -623,7 +653,7 @@ class Reformer(nn.Module):
 
             blocks.append(nn.ModuleList([f, g]))
 
-        self.layers = ReversibleSequence(nn.ModuleList(blocks), layer_dropout = layer_dropout, reverse_thres = reverse_thres)
+        self.layers = ReversibleSequence(nn.ModuleList(blocks), layer_dropout = layer_dropout, reverse_thres = reverse_thres, send_signal = True)
 
     def forward(self, x, **kwargs):
         x = torch.cat([x, x], dim = -1)
@@ -632,7 +662,7 @@ class Reformer(nn.Module):
         return torch.stack(x.chunk(2, dim=-1)).sum(dim=0)
 
 class ReformerLM(nn.Module):
-    def __init__(self, num_tokens, dim, depth, max_seq_len, heads = 8, bucket_size = 64, n_hashes = 4, add_local_attn_hash = False, ff_chunks = 100, attn_chunks = 1, causal = False, weight_tie = False, lsh_dropout = 0., ff_dropout = 0., ff_mult = 4, ff_activation = None, post_attn_dropout = 0., layer_dropout = 0., random_rotations_per_head = False, twin_attention = False, use_scale_norm = False, use_full_attn = False, full_attn_thres = 0, reverse_thres = 0, num_mem_kv = 0, one_value_head = False, emb_dim = None, return_embeddings = False, weight_tie_embedding = False, fixed_position_emb = False, axial_position_emb = False, axial_position_shape = (), axial_position_dims = ()):
+    def __init__(self, num_tokens, dim, depth, max_seq_len, heads = 8, bucket_size = 64, n_hashes = 4, ff_chunks = 100, attn_chunks = 1, causal = False, weight_tie = False, lsh_dropout = 0., ff_dropout = 0., ff_mult = 4, ff_activation = None, ff_glu = False, post_attn_dropout = 0., layer_dropout = 0., random_rotations_per_head = False, twin_attention = False, use_scale_norm = False, use_full_attn = False, full_attn_thres = 0, reverse_thres = 0, num_mem_kv = 0, one_value_head = False, emb_dim = None, return_embeddings = False, weight_tie_embedding = False, fixed_position_emb = False, axial_position_emb = False, axial_position_shape = (), axial_position_dims = ()):
         super().__init__()
         emb_dim = default(emb_dim, dim)
         self.max_seq_len = max_seq_len
@@ -649,7 +679,7 @@ class ReformerLM(nn.Module):
         else:
             self.pos_emb = AbsolutePositionalEmbedding(emb_dim, max_seq_len)
 
-        self.reformer = Reformer(dim, depth, max_seq_len, heads = heads, bucket_size = bucket_size, n_hashes = n_hashes, add_local_attn_hash = add_local_attn_hash, ff_chunks = ff_chunks, attn_chunks = attn_chunks, causal = causal, weight_tie = weight_tie, lsh_dropout = lsh_dropout, ff_mult = ff_mult, ff_activation = ff_activation, ff_dropout = ff_dropout, post_attn_dropout = 0., layer_dropout = layer_dropout, random_rotations_per_head = random_rotations_per_head, twin_attention = twin_attention, use_scale_norm = use_scale_norm, use_full_attn = use_full_attn, full_attn_thres = full_attn_thres, reverse_thres = reverse_thres, num_mem_kv = num_mem_kv, one_value_head = one_value_head)
+        self.reformer = Reformer(dim, depth, max_seq_len, heads = heads, bucket_size = bucket_size, n_hashes = n_hashes, ff_chunks = ff_chunks, attn_chunks = attn_chunks, causal = causal, weight_tie = weight_tie, lsh_dropout = lsh_dropout, ff_mult = ff_mult, ff_activation = ff_activation, ff_glu = ff_glu, ff_dropout = ff_dropout, post_attn_dropout = 0., layer_dropout = layer_dropout, random_rotations_per_head = random_rotations_per_head, twin_attention = twin_attention, use_scale_norm = use_scale_norm, use_full_attn = use_full_attn, full_attn_thres = full_attn_thres, reverse_thres = reverse_thres, num_mem_kv = num_mem_kv, one_value_head = one_value_head)
 
         if return_embeddings:
             self.out = Identity()
